from __future__ import annotations
import typer

from svc_infra.db.cli import app as _db_app
from svc_infra.auth.cli import app as _auth_app

app = typer.Typer(
    name="svc-infra",
    help="Unified CLI for service infrastructure commands (auth, db, agent, etc).",
    no_args_is_help=True,
    add_completion=False
)
app.add_typer(_db_app, name="db", help="Database related commands")
app.add_typer(_auth_app, name="auth", help="Auth db setup related commands")

if __name__ == "__main__":
    app()


from __future__ import annotations


# Lazy exports to avoid importing optional or heavy submodules at package import time.
# Accessing these attributes will import the underlying implementations on demand.

def __getattr__(name: str):
    if name == "include_auth":
        from .include_auth import include_auth as _f

        return _f
    if name == "get_fastapi_users":
        from .users import get_fastapi_users as _f

        return _f
    if name == "oauth_router":
        from .oauth_router import oauth_router as _f

        return _f
    if name == "get_auth_settings":
        # This module may not exist in consumer projects until scaffolded.
        from .settings import get_auth_settings as _f  # type: ignore

        return _f
    raise AttributeError(f"module 'svc_infra.auth' has no attribute {name!r}")



from fastapi import APIRouter, Response, status
import logging

router = APIRouter()

@router.get("/ping", status_code=status.HTTP_200_OK)
def ping():
    logging.info("Health check: /ping endpoint accessed. Service is responsive.")
    return Response(status_code=status.HTTP_200_OK)



from __future__ import annotations

from pathlib import Path
import typer

from .core import (
    scaffold_auth_core,
    scaffold_auth_models_core,
    scaffold_auth_schemas_core,
)

app = typer.Typer(no_args_is_help=True, add_completion=False)

def _echo_result(r: dict) -> None:
    action = r.get("action")
    path = r.get("path")
    if action == "wrote":
        typer.echo(f"Wrote {path}")
    elif action == "skipped":
        reason = r.get("reason", "skipped")
        typer.echo(f"SKIP {path} ({reason}). Use --overwrite to replace.")
    else:
        typer.echo(f"{action or 'done'} {path}")

@app.command("scaffold-auth", help="Scaffold auth models.py and schemas.py from templates")
def scaffold_auth(
        models_dir: Path = typer.Option(..., help="Where to place models.py"),
        schemas_dir: Path = typer.Option(..., help="Where to place schemas.py"),
        overwrite: bool = typer.Option(False, help="Overwrite files if they exist"),
):
    res = scaffold_auth_core(models_dir=models_dir, schemas_dir=schemas_dir, overwrite=overwrite)
    _echo_result(res["results"]["models"])
    _echo_result(res["results"]["schemas"])

@app.command("scaffold-auth-models", help="Scaffold auth models.py from template")
def scaffold_auth_models(
        dest_dir: Path = typer.Option(..., help="Directory to place models.py"),
        overwrite: bool = typer.Option(False, help="Overwrite if exists"),
):
    res = scaffold_auth_models_core(dest_dir=dest_dir, overwrite=overwrite)
    _echo_result(res["result"])

@app.command("scaffold-auth-schemas", help="Scaffold auth schemas.py from template")
def scaffold_auth_schemas(
        dest_dir: Path = typer.Option(..., help="Directory to place schemas.py"),
        overwrite: bool = typer.Option(False, help="Overwrite if exists"),
):
    res = scaffold_auth_schemas_core(dest_dir=dest_dir, overwrite=overwrite)
    _echo_result(res["result"])


from __future__ import annotations
from typing import Dict, Any
from fastapi import APIRouter, Request, HTTPException
from fastapi.responses import RedirectResponse
from authlib.integrations.starlette_client import OAuth
from sqlalchemy import select
from fastapi_users.authentication import AuthenticationBackend
from fastapi_users.password import PasswordHelper

from svc_infra.api.fastapi.db.integration import SessionDep

def oauth_router_with_backend(
        user_model: type,
        auth_backend: AuthenticationBackend,
        providers: Dict[str, Dict[str, Any]],
        post_login_redirect: str = "/",
        prefix: str = "/auth/oauth",
) -> APIRouter:
    oauth = OAuth()

    # Register all providers
    for name, cfg in providers.items():
        kind = cfg.get("kind")
        if kind == "oidc":
            oauth.register(
                name,
                client_id=cfg["client_id"],
                client_secret=cfg["client_secret"],
                server_metadata_url=f"{cfg['issuer'].rstrip('/')}/.well-known/openid-configuration",
                client_kwargs={"scope": cfg.get("scope", "openid email profile")},
            )
        elif kind in ("github", "linkedin"):
            oauth.register(
                name,
                client_id=cfg["client_id"],
                client_secret=cfg["client_secret"],
                authorize_url=cfg["authorize_url"],
                access_token_url=cfg["access_token_url"],
                api_base_url=cfg["api_base_url"],
                client_kwargs={"scope": cfg.get("scope", "")},
            )
        else:
            # you can add more branches (facebook, apple*) as needed
            pass

    r = APIRouter(prefix=prefix, tags=["auth:oauth"])

    @r.get("/{provider}/login")
    async def oauth_login(request: Request, provider: str):
        client = oauth.create_client(provider)
        if not client:
            raise HTTPException(404, "Provider not configured")
        redirect_uri = request.url_for("oauth_callback", provider=provider)
        return await client.authorize_redirect(request, str(redirect_uri))

    @r.get("/{provider}/callback", name="oauth_callback")
    async def oauth_callback(request: Request, provider: str, session: SessionDep):
        client = oauth.create_client(provider)
        if not client:
            raise HTTPException(404, "Provider not configured")

        token = await client.authorize_access_token(request)

        email = None
        full_name = None

        cfg = providers.get(provider, {})
        kind = cfg.get("kind")

        if kind == "oidc":
            userinfo = token.get("userinfo") or await client.parse_id_token(request, token)
            email = userinfo.get("email")
            full_name = userinfo.get("name") or userinfo.get("preferred_username")
        elif kind == "github":
            resp = await client.get("user", token=token)
            data = resp.json()
            email = data.get("email")
            if not email:
                emails = (await client.get("user/emails", token=token)).json()
                primary = next((e for e in emails if e.get("primary")), emails[0] if emails else {})
                email = primary.get("email")
            full_name = data.get("name") or data.get("login")
        elif kind == "linkedin":
            # profile
            me = (await client.get("me", token=token)).json()
            # email
            em = (await client.get(
                "emailAddress?q=members&projection=(elements*(handle~))",
                token=token
            )).json()
            elements = em.get("elements") or []
            if elements and "handle~" in elements[0]:
                email = elements[0]["handle~"].get("emailAddress")
            localizedFirst = (((me.get("firstName") or {}).get("localized")) or {}).values()
            localizedLast = (((me.get("lastName") or {}).get("localized")) or {}).values()
            first = next(iter(localizedFirst), None)
            last = next(iter(localizedLast), None)
            full_name = " ".join([x for x in [first, last] if x])
        else:
            raise HTTPException(400, "Unsupported provider kind")

        if not email:
            raise HTTPException(400, "No email from provider")

        # Upsert user
        existing = (await session.execute(select(user_model).filter_by(email=email))).scalars().first()
        if existing:
            user = existing
        else:
            user = user_model(email=email, is_active=True, is_superuser=False, is_verified=True)
            user.hashed_password = PasswordHelper().hash("!oauth!")
            if hasattr(user, "full_name"):
                setattr(user, "full_name", full_name)
            session.add(user)
            await session.flush()

        jwt = (auth_backend.get_strategy)().write_token(user)
        return RedirectResponse(url=f"{post_login_redirect}?token={jwt}")

    return r


import inspect
import asyncio
from functools import wraps

def _async_cmd(fn):
    sig = inspect.signature(fn)
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return asyncio.run(fn(*args, **kwargs))
    wrapper.__signature__ = sig
    return wrapper


from __future__ import annotations

import os
import warnings
from enum import StrEnum
from functools import cache
from typing import NamedTuple


class Environment(StrEnum):
    LOCAL = "local"
    DEV   = "dev"
    TEST  = "test"
    PROD  = "prod"


# Map common aliases -> canonical
SYNONYMS: dict[str, Environment] = {
    "development": Environment.DEV,
    "dev": Environment.DEV,
    "local": Environment.LOCAL,
    "test": Environment.TEST,
    "preview": Environment.TEST,
    "prod": Environment.PROD,
    "production": Environment.PROD,
    "staging": Environment.TEST,  # Treat 'staging' as 'test' for environment purposes
}

ALL_ENVIRONMENTS = {e for e in Environment}


def _normalize(raw: str | None) -> Environment | None:
    """
    Normalize raw environment string to canonical Env enum, case-insensitively.
    """
    if not raw:
        return None
    val = raw.strip().casefold()  # case-insensitive, handles unicode edge cases
    # Check against canonical enum values
    if val in (e.value for e in Environment):
        return Environment(val)  # exact match
    # Check against synonyms
    return SYNONYMS.get(val)


@cache
def get_current_environment() -> Environment:
    """
    Resolve the current environment once, with sensible fallbacks.

    Precedence:
      1) APP_ENV
      2) RAILWAY_ENVIRONMENT_NAME
      3) "local" (default)

    Unknown values fall back to LOCAL with a one-time warning.
    """
    raw = os.getenv("APP_ENV") or os.getenv("RAILWAY_ENVIRONMENT_NAME")
    env = _normalize(raw)
    if env is None:
        if raw:
            warnings.warn(
                f"Unrecognized environment '{raw}', defaulting to 'local'.",
                RuntimeWarning,
                stacklevel=2,
            )
        env = Environment.LOCAL
    return env


class EnvironmentFlags(NamedTuple):
    environment: Environment
    is_local: bool
    is_dev: bool
    is_test: bool
    is_prod: bool


def get_environment_flags(environment: Environment | None = None) -> EnvironmentFlags:
    e = environment or get_current_environment()
    return EnvironmentFlags(
        environment=e,
        is_local=(e == _normalize("local")),
        is_dev=(e == _normalize("dev")),
        is_test=(e == _normalize("test")),
        is_prod=(e == _normalize("prod")),
    )


# Handy globals
CURRENT_ENVIRONMENT: Environment = get_current_environment()
ENV_FLAGS: EnvironmentFlags = get_environment_flags(CURRENT_ENVIRONMENT)
IS_LOCAL, IS_DEV, IS_TEST, IS_PROD = ENV_FLAGS.is_local, ENV_FLAGS.is_dev, ENV_FLAGS.is_test, ENV_FLAGS.is_prod


def pick(*, prod, nonprod=None, dev=None, test=None, local=None):
    """
    Choose a value based on the active environment.

    Example:
        log_level = pick(prod="INFO", nonprod="DEBUG", dev="DEBUG")
    """
    e = get_current_environment()
    if e is Environment.PROD:
        return prod
    if e is Environment.DEV and dev is not None:
        return dev
    if e is Environment.TEST and test is not None:
        return test
    if e is Environment.LOCAL and local is not None:
        return local
    if nonprod is not None:
        return nonprod
    raise ValueError("pick(): No value found for environment and 'nonprod' was not provided.")



import logging
import traceback
from fastapi import Request
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError, HTTPException
from sqlalchemy.exc import SQLAlchemyError
from starlette.exceptions import HTTPException as StarletteHTTPException

from svc_infra.app import IS_PROD
from svc_infra.api.fastapi.middleware.errors.exceptions import FastApiException

logger = logging.getLogger(__name__)

def get_error_type(exc: Exception) -> str:
    return getattr(exc, 'error', None) or type(exc).__name__

def log_exception(level: int, exc: Exception, request: Request, status_code: int) -> None:
    logger.log(
        level,
        f"{request.method} {request.url.path} [{status_code}] {type(exc).__name__}: {exc}",
        exc_info=True
    )

def format_error_response(
    exc: Exception,
    request: Request,
    status_code: int,
    detail: str | dict | list
) -> JSONResponse:
    error_type = get_error_type(exc)
    log_exception(logging.ERROR, exc, request, status_code)

    response_content = {
        "error": error_type,
        "detail": detail if not IS_PROD else (
            "Something went wrong. Please contact support."
            if status_code == 500 else detail
        ),
    }

    if not IS_PROD and status_code == 500:
        response_content["trace"] = traceback.format_exc()

    return JSONResponse(status_code=status_code, content=response_content)

def register_error_handlers(app):
    @app.exception_handler(FastApiException)
    async def handle_api_frameworks_exception(request: Request, exc: FastApiException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(RequestValidationError)
    async def handle_validation_error(request: Request, exc: RequestValidationError):
        detail = exc.errors() if not IS_PROD else "Invalid request payload."
        return format_error_response(exc, request, 422, detail)

    @app.exception_handler(HTTPException)
    async def handle_http_exception(request: Request, exc: HTTPException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(StarletteHTTPException)
    async def handle_starlette_http_exception(request: Request, exc: StarletteHTTPException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(SQLAlchemyError)
    async def handle_sqlalchemy_error(request: Request, exc: SQLAlchemyError):
        return format_error_response(exc, request, 500, "Please try again later.")

    @app.exception_handler(Exception)
    async def handle_unexpected_error(request: Request, exc: Exception):
        return format_error_response(exc, request, 500, str(exc))
